__version__ = "0.4.0rc3"

from .core.error.error import Error
from .tools.not_found_error import NotFoundError
from .core.view.view import View
from .core.assembler.build import Build
from .core.emt.emt import Emt
from .tools.get_mode import get_mode
from .tools.get_root_path import get_root_path
from .core.sv.sv import Sv
from .core.app.puft import Puft
from .core.db.db import Db, orm
from .core.cell.cell import Cell
from .core.sv.sv_cell import SvCell
from .core.emt.emt_cell import EmtCell
from .core.view.view_cell import ViewCell
from .core.error.error_cell import ErrorCell
from .core.test.test import Test
from .core.test.test_mock import Mock
from .tools.login_required_dec import login_required
from .tools.log import log
from .core.db.model_not_found_error import ModelNotFoundError
from .core.sock.socket import Socket
from .core.sock.sock_cell import SockCell
from .core.sock.sock import Sock



from __future__ import annotations
import os
from typing import Dict, TYPE_CHECKING, Any

from warepy import (
    join_paths, format_message, load_yaml, get_enum_values, Singleton
)
from flask_socketio import SocketIO

from puft.core.sock.socket import Socket
from puft.tools.hints import CLIModeEnumUnion
from puft.tools.log import log
from puft.core.app.app_mode_enum import AppModeEnum
from puft.core.cli.cli_run_enum import CLIRunEnum
from puft.core.error.error import Error
from puft.tools.error_handlers import handle_wildcard_error
from puft.core.cell.named_cell import NamedCell
from puft.core.cell.config_cell import ConfigCell
from puft.core.app.puft_sv_cell import PuftSvCell
from puft.core.db.db_sv_cell import DbSvCell
from puft.core.sock.sock_sv_cell import SocketSvCell
from puft.core.sv.sv_cell import SvCell
from puft.core.view.view_cell import ViewCell
from puft.core.emt.emt_cell import EmtCell
from puft.core.error.error_cell import ErrorCell

from puft.core.db.db import Db
from puft.core.app.puft import Puft
from puft.tools.hints import CLIModeEnumUnion
from .config_extension_enum import ConfigExtensionEnum

if TYPE_CHECKING:
    from .build import Build


class Assembler(Singleton):
    """Assembles all project instances from given `Builder` type's class and
    initializes it.
    
    Acts automatically and shouldn't be inherited directly by project in any
    form.
    """
    DEFAULT_LOG_PARAMS = {
        "path": "./var/logs/system.log",
        "level": "DEBUG",
        "format":
            "{time:%Y.%m.%d at %H:%M:%S:%f%z} | {level} | {extra} >> {message}",
        "rotation": "10 MB",
        "serialize": False
    }

    def __init__(
        self, 
        mode_enum: CLIModeEnumUnion, host: str, port: int,
        build: Build,
        *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        # Define attributes for getter methods to be used at builder.
        # Do not set extra_configs_by_name to None at initialization, because
        # `get()` method called from this dictionary.
        self.extra_configs_by_name = {}
        self.root_path = build.root_path
        self.sv_cells = build.sv_cells
        self.view_cells = build.view_cells
        self.error_cells: list[ErrorCell] = build.error_cells
        self.emt_cells = build.emt_cells
        self.mode_enum = mode_enum
        self.shell_processors = build.shell_processors
        self.cli_cmds = build.cli_cmds
        self.ctx_processor_func = build.ctx_processor_func
        self.each_request_func = build.each_request_func
        self.first_request_func = build.first_request_func
        self.default_wildcard_error_handler_func = handle_wildcard_error
        self.socket_handler_cells = build.sock_handler_cells
        self.default_sock_error_handler = build.default_sock_error_handler

        self.socket_enabled: bool = False

        self._assign_config_cells(build.config_dir)

        # Traverse given configs and assign enabled builtin cells.
        self._assign_builtin_sv_cells(mode_enum, host, port)

    @log.catch
    def get_puft(self) -> Puft:
        return self.puft

    @log.catch
    def get_db(self) -> Db:
        return self.db

    @log.catch
    def get_root_path(self) -> str:
        return self.root_path

    @log.catch
    def get_mode_enum(self) -> CLIModeEnumUnion:
        return self.mode_enum

    @log.catch
    def _assign_config_cells(self, config_dir: str) -> None:
        """Traverse through config files under given config_dir and create 
        ConfigCells from them.

        Name taken from filename of config and should be the same as specified 
        at config's target sv_cell.name.

        Names can contain additional extension like `name.prod.yaml` according
        to appropriate Puft modes. These configs launched per each mode. Config
        without this extra extension, considered `prod` moded.
        """
        self.config_cells: list[ConfigCell] = []

        config_path: str = join_paths(self.root_path, config_dir)
        source_map_by_name: dict[str, dict[AppModeEnum, str]] = \
            self._find_config_files(config_path)

        for name, source_map in source_map_by_name.items():
            self.config_cells.append(ConfigCell(
                name=name,
                source_by_app_mode=source_map))

    def _find_config_files(
            self, config_path: str) -> dict[str, dict[AppModeEnum, str]]:
        """Accept path to config dir and return dict describing all paths to
        configs for all app modes per sv name.
        
        Return example:
        ```python
        {
            "custom_sv_name": {
                AppModeEnum.PROD: "./some/path",
                AppModeEnum.DEV: "./some/path",
                AppModeEnum.TEST: "./some/path"
            }
        }
        ```
        """
        source_map_by_name = {} 
        for filename in os.listdir(config_path):
            # Pick files only under config_dir.
            if os.path.isfile(join_paths(config_path, filename)):
                parts = filename.split(".")
                file_path = join_paths(config_path, filename)

                if len(parts) == 1:
                    # Skip files without extension.
                    continue
                # Check if file has supported extension.
                elif len(parts) == 2:
                    # Config name shouldn't contain dots and thus we can grab
                    # it right here.
                    config_name = parts[0]
                    if parts[1] in get_enum_values(ConfigExtensionEnum):
                        # Add file without app mode automatically to
                        # `prod`.
                        if config_name not in source_map_by_name:
                            source_map_by_name[config_name] = dict()
                        source_map_by_name[config_name][
                            AppModeEnum.PROD] = file_path
                    else:
                        # Skip files with unsupported extension.
                        continue
                elif len(parts) == 3:
                    # Config name shouldn't contain dots and thus we can grab
                    # it right here.
                    config_name = parts[0]
                    if parts[1] in get_enum_values(AppModeEnum) \
                            and parts[2] in get_enum_values(
                                ConfigExtensionEnum):
                        # File has both normal extension and defined
                        # app mode.
                        if config_name not in source_map_by_name:
                            source_map_by_name[config_name] = dict()
                        source_map_by_name[config_name][
                            AppModeEnum(parts[1])] = file_path
                    else:
                        # Unrecognized app mode or extension,
                        # maybe raise warning?
                        continue
                else:
                    # Skip files with names containing dots, e.g.
                    # "dummy.new.prod.yaml".
                    continue
        return source_map_by_name

    @log.catch
    def _assign_builtin_sv_cells(
            self, mode_enum: CLIModeEnumUnion, host: str, port: int) -> None:
        """Assign builting sv cells if configuration file for its sv
        exists.
        """
        self.builtin_sv_cells: list[Any] = [PuftSvCell(
            name="puft",
            sv_class=Puft,
            mode_enum=mode_enum,
            host=host,
            port=port
        )]
        log_layers: list[str] = []

        # Enable only modules with specified configs.
        if self.config_cells:
            try:
                NamedCell.find_by_name("db", self.config_cells)
            except ValueError:
                pass
            else:
                self.builtin_sv_cells.append(DbSvCell(
                    name="db",
                    sv_class=Db
                ))
                log_layers.append('db')

            try:
                NamedCell.find_by_name('sock', self.config_cells)
            except ValueError:
                pass
            else:
                self.builtin_sv_cells.append(SocketSvCell(
                    name='socket',
                    sv_class=Socket))
                self.socket_enabled = True
                log_layers.append('socket')
            
            if log_layers:
                log.info(f'Enabled layers: {", ".join(log_layers)}')

    @staticmethod
    @log.catch
    def build(
            configs_by_name: Dict[str, dict] | None = None,
            root_path: str | None = None) -> None:
        """Initialize all given to Assembler instances in their dependencies.
        
        Reassign assembler's attributes to given ones and run it's setup
        operations to assemble app and other project's instances.

        Args:
            configs_by_name (optional):
                Configs to be appended to appropriate ones described in Build
                class. Defaults to None.
                Contain config name as key (which is compared to ConfigCell
                name field) and configuration mapping as value, e.g.:
        ```python
        configs_by_name = {
            "app": {"TESTING": True},
            "db": {"db_uri": "sqlite3:///:memory:"}
        }
        ```
            root_path (optional):
                Root path to execute project from. Defaults to `os.getcwd()`.
                Required in cases of calling this function not from actual
                project root (e.g. from tests) to set root path explicitly.
        """
        # Get Assembler instance without args, because it should be initialized before (in root create_app function).
        assembler = Assembler.instance()

        if root_path:
            assembler.root_path = root_path
        if configs_by_name:
            assembler.extra_configs_by_name = configs_by_name

        assembler._build_all()
        
    @log.catch
    def _build_all(self) -> None:
        """Send commands to build all given instances."""
        self._build_log()
        self._build_svs()
        self._build_views()
        self._build_errors()
        self._build_emts()
        self._build_shell_processors()
        self._build_cli_cmds()
        self._build_sock_handlers()

        # Call postponed build from created App.
        try:
            self.puft.postbuild()
        except NotImplementedError:
            pass

    @log.catch
    def _build_log(self) -> None:
        """Call chain to build log."""
        # Try to find log config cell and build log class from it
        if self.config_cells:
            try:
                log_config_cell = NamedCell.find_by_name("log", self.config_cells)
            except ValueError:
                log_config = None
            else:
                # Parse config mapping from cell and append extra configs,
                # if they are given
                app_mode_enum: AppModeEnum
                if type(self.mode_enum) is CLIRunEnum:
                    app_mode_enum = AppModeEnum(self.mode_enum.value) 
                else:
                    # Assign dev app mode for all other app modes
                    app_mode_enum = AppModeEnum.DEV
                log_config = log_config_cell.parse(
                    app_mode_enum=app_mode_enum,
                    root_path=self.root_path, 
                    update_with=self.extra_configs_by_name.get("log", None)
                )
            self._init_log_class(config=log_config)

    @log.catch
    def _init_log_class(self, config: dict | None = None) -> None:
        """Build log with given config.
        
        If config is None, build with default parameters."""
        # Use full or partially (with replacing missing keys from default) given config.
        log_kwargs = self.DEFAULT_LOG_PARAMS
        if config:
            for k, v in config.items():
                log_kwargs[k] = v
        log.configure(**log_kwargs)

    @log.catch
    def _build_svs(self) -> None:
        self._run_builtin_sv_cells()
        self._run_custom_sv_cells()

    @log.catch
    def _build_sock_handlers(self) -> None:
        if self.socket_handler_cells and self.socket_enabled:
            for cell in self.socket_handler_cells:
                socket: SocketIO = self.socket.get_socket()

                # Register class for socketio namespace
                # https://flask-socketio.readthedocs.io/en/latest/getting_started.html#class-based-namespaces
                socket.on_namespace(cell.handler_class(cell.namespace))
                # Also register error handler for the same namespace
                socket.on_error(cell.namespace)(cell.error_handler) 

    @log.catch
    def _perform_db_postponed_setup(self) -> None:
        """Postponed setup is required, because Db uses Flask app to init
        native SQLAlchemy db inside, so it's possible only after App
        initialization.

        The setup_db requires native flask app to work with.
        """
        self.db.setup(flask_app=self.puft.get_native_app())
    
    @log.catch
    def _run_builtin_sv_cells(self) -> None:
        for cell in self.builtin_sv_cells:
            # Check for domain's config in given cells by comparing names and
            # apply to sv config if it exists
            config = self._assemble_sv_config(name=cell.name) 

            # Each builtin sv should receive essential fields for their
            # configs, such as root_path, because they cannot import Assembler
            # due to circular import issue and get this fields by themselves
            config["root_path"] = self.root_path

            # Initialize sv.
            if type(cell) is PuftSvCell:
                # Run special initialization with mode, host and port for Puft
                # sv
                self.puft: Puft = cell.sv_class(
                    mode_enum=cell.mode_enum, host=cell.host, port=cell.port, 
                    config=config,
                    ctx_processor_func=self.ctx_processor_func,
                    each_request_func=self.each_request_func,
                    first_request_func=self.first_request_func)
            elif type(cell) is DbSvCell:
                self.db: Db = cell.sv_class(config=config)
                # Perform Db postponed setup
                self._perform_db_postponed_setup()
            elif type(cell) is SocketSvCell:
                self.socket = cell.sv_class(config=config, app=self.puft)
            else:
                cell.sv_class(config=config)

    @log.catch
    def _run_custom_sv_cells(self) -> None:
        if self.sv_cells:
            for cell in self.sv_cells:
                if self.config_cells:
                    sv_config = self._assemble_sv_config(name=cell.name) 
                else:
                    sv_config = {}

                cell.sv_class(config=sv_config)

    @log.catch
    def _assemble_sv_config(
            self,
            name: str, is_errors_enabled: bool = False) -> dict[str, Any]:
        """Check for sv's config in config cells by comparing its given
        name and return it as dict.

        If appropriate config hasn't been found, raise ValueError if
        `is_errors_enabled = True` or return empty dict otherwise.
        """
        try:
            config_cell_with_target_name: ConfigCell = NamedCell.find_by_name(
                name, self.config_cells)
        except ValueError:
            # If config not found and errors enabled, raise error.
            if is_errors_enabled:
                message = format_message(
                    "Appropriate config for given name {} hasn't been found.",
                    name)
                raise ValueError(message)
            else:
                config: dict[str, Any] = {}
        else:
            if type(config_cell_with_target_name) is not ConfigCell:
                raise TypeError(format_message(
                    "Type of cell should be ConfigCell, but {} received",
                    type(config_cell_with_target_name)))
            else:
                app_mode_enum: AppModeEnum
                if type(self.mode_enum) is CLIRunEnum:
                    app_mode_enum = AppModeEnum(self.mode_enum.value) 
                else:
                    # Assign dev app mode for all other app modes.
                    app_mode_enum = AppModeEnum.DEV
                config = config_cell_with_target_name.parse(
                    root_path=self.root_path, 
                    update_with=self.extra_configs_by_name.get(name, None),
                    app_mode_enum=app_mode_enum)
        return config

    @log.catch
    def _fetch_yaml_project_version(self) -> str:
        """Fetch project version from info.yaml from the root path and return it. 

        TODO:
            Replace this logic with senseful one. Better use configuration's
            version? Or __init__.py or main.py specified.
        """
        info_data = load_yaml(join_paths(self.root_path, "./info.yaml"))
        project_version = info_data["version"]
        return project_version

    @log.catch
    def _build_views(self) -> None:
        """Build all views by registering them to app."""
        if self.view_cells:
            for view_cell in self.view_cells:
                self.puft.register_view(view_cell)

    @log.catch
    def _build_emts(self) -> None:
        """Build emts from given cells and inject Puft application controllers to each."""
        if self.emt_cells:
            for cell in self.emt_cells:
                cell.emt_class(puft=self.puft)

    @log.catch
    def _build_shell_processors(self) -> None:
        if self.shell_processors:
            self.puft.register_shell_processor(*self.shell_processors)

    @log.catch
    def _build_cli_cmds(self) -> None:
        if self.cli_cmds:
            self.puft.register_cli_cmd(*self.cli_cmds)

    def _build_errors(self) -> None:
        # TODO: Test case when user same error class registered twice (e.g. in
        # duplicate cells)
        is_wildcard_specified = False
        for error_cell in self.error_cells:
            if type(error_cell.error_class) is Error:
                is_wildcard_specified = True
            self.puft.register_error(
                error_cell.error_class, error_cell.handler_function)
        # If wildcard handler is not specified, apply the default one
        if not is_wildcard_specified:
            self.puft.register_error(
                Error, self.default_wildcard_error_handler_func)



class Error(Exception):
    """Main error class to be inherited in all app's custom errors."""
    def __init__(self, message: str = '', status_code: int = 400) -> None:
        super().__init__(message)
        if type(message) is str:
            self.message: str = message
        else:
            raise TypeError(
                'Error first argument (message) should be str,'
                f' got {type(message)} instead')
        if type(status_code) is int:
            self.status_code: int = status_code
        else:
            raise TypeError(
                'Error second argument (status_code) should be int,'
                f' got {type(status_code)} instead')

    def expose(self) -> dict:
        return {"error": {
            "name": self.__class__.__name__,
            "message": "; ".join(self.args),
            "status_code": self.status_code
        }}



import re
import os
import json
from copy import copy
from dataclasses import dataclass
from typing import Any

from warepy import join_paths, load_yaml

from puft.core.app.app_mode_enum import AppModeEnum
from ..assembler.config_extension_enum import ConfigExtensionEnum
from .named_cell import NamedCell

@dataclass
class ConfigCell(NamedCell):
    """Config cell which can be used to load configs to appropriate instance's
    configuration by name."""
    source_by_app_mode: dict[AppModeEnum, str]

    def parse(
            self, app_mode_enum: AppModeEnum, root_path: str,
            update_with: dict[str, Any] | None = None,
            convert_keys_to_lower: bool = True) -> dict[str, Any]:
        """Parse config cell and return configuration dictionary.

        Args:
            app_mode_enum:
                App mode to run appropriate config.
            root_path:
                Path to join config cell source with.
            update_with (optional):
                Dictionary to update config cell mapping with.
                Defaults to None.
            convert_keys_to_lower (optional):
                If true, all keys from origin mapping and mapping from
                `update_with` will be converted to upper case.
        
        Raise:
            ValueError:
                If given config cell's source has unrecognized extension.
        """
        res_config: dict[str, Any] = {}

        config_by_mode: dict[AppModeEnum, dict] = self._load_config_by_mode()
        res_config = self._update_config_for_mode(config_by_mode, app_mode_enum)

        if res_config:
            self._parse_string_config_values(res_config, root_path)
        else:
            res_config = {}

        # Update given config with extra dictionary if this dictionary given
        # and not empty.
        if update_with:
            res_config.update(update_with)

        if convert_keys_to_lower:
            temp_config = {}
            for k, v in res_config.items():
                temp_config[k.lower()] = v
            res_config = temp_config

        return res_config

    def _update_config_for_mode(
            self,
            config_by_mode: dict[AppModeEnum, dict],
            app_mode_enum: AppModeEnum) -> dict:
        """Take config maps for each mode and return result config updated for
        current mode.
        
        E.g., given mode is TEST, so final config will be PROD config updated
        by DEV config and then updated by TEST config (so test keys will
        take priority).
        """ 
        prod_config = copy(config_by_mode[AppModeEnum.PROD])

        if app_mode_enum is AppModeEnum.TEST:
            dev_config = copy(config_by_mode[AppModeEnum.DEV])
            test_config = copy(config_by_mode[AppModeEnum.TEST])
            dev_config.update(test_config)
            prod_config.update(dev_config)
        elif app_mode_enum is AppModeEnum.DEV:
            dev_config = copy(config_by_mode[AppModeEnum.DEV])
            prod_config.update(dev_config)
        else:
            # Prod mode, do nothing extra
            pass
        return prod_config
    
    def _load_config_by_mode(self) -> dict[AppModeEnum, dict]:
        config_by_mode: dict[AppModeEnum, dict] = {}
        for app_mode_enum in AppModeEnum:
            try:
                source = self.source_by_app_mode[app_mode_enum]
            except KeyError:
                # No source for such mode
                config_by_mode[app_mode_enum] = {}
                continue
            source_extension = source[source.rfind(".")+1:]
            config_by_mode[app_mode_enum] = self._load_config_from_file(
                source_extension_enum=ConfigExtensionEnum(source_extension),
                source=source)
        return config_by_mode
    
    def _load_config_from_file(
            self,
            source_extension_enum: ConfigExtensionEnum, source: str) -> dict:
        # Fetch extension and load config from file.
        match source_extension_enum:
            case ConfigExtensionEnum.JSON:
                with open(source, "r") as config_file:
                    config = json.load(config_file)
            case ConfigExtensionEnum.YAML:
                config = load_yaml(source)
                if config is None:
                    raise NotImplementedError(self.name)
            case _:
                raise ValueError("Unrecognized config cell source's extension")
        return config
    
    def _parse_string_config_values(
            self, config: dict[str, Any], root_path: str) -> None:
        for k, v in config.items():
            if type(v) == str:
                # Find environs to be requested
                # Exclude escaped curly brace like `\{not_environ}`
                # Note that environs matching \w+ pattern only supported
                envs: list[str] = re.findall(r"[^\\]\{\w+\}", v)
                if envs:
                    for env in [
                                x.replace("{", "").replace("}", "")
                                for x in envs
                            ]:
                        env = env.strip()
                        real_env_value = os.getenv(env.strip())
                        if real_env_value is None:
                            raise ValueError(
                                f"Environ {env} specified in field"
                                f" {self.name}.{k} was not found")
                        else:
                            v = v.replace("{" + f"{env}" + "}", real_env_value)

                # Look for escaped curly braces and normalize them.
                v = v.replace(r"\{", "{").replace(r"\}", "}")

                # Find paths required to be joined to the root path.
                if v[0] == "." and v[1] == "/":
                    config[k] = join_paths(root_path, v)
                else:
                    config[k] = v


from dataclasses import dataclass
from typing import Callable

from puft.core.cell.cell import Cell
from .sock import Sock
from .default_sock_error_handler import default_sock_error_handler


@dataclass
class SockCell(Cell):
    namespace: str
    handler_class: type[Sock]
    error_handler: Callable = default_sock_error_handler



from dataclasses import dataclass

from puft.core.cell.cell import Cell

from .view import View


@dataclass
class ViewCell(Cell):
    endpoint: str
    view_class: type[View]
    route: str  # Route will be the same for all methods.


from puft.core.cli.cli_db_enum import CLIDbEnum
from puft.core.cli.cli_run_enum import CLIRunEnum
from puft.core.cli.cli_helper_enum import CLIHelperEnum

CLIModeEnumUnion = CLIDbEnum | CLIRunEnum | CLIHelperEnum



from typing import TYPE_CHECKING

from flask_socketio import (
    SocketIO, send, emit, join_room, leave_room, SocketIOTestClient)

from puft.core.sv.sv import Sv

if TYPE_CHECKING:
    from puft.core.app.puft import Puft


class Socket(Sv):
    def __init__(self, config: dict, app: 'Puft') -> None:
        super().__init__(config)
        self.app = app
        self.socket: SocketIO = SocketIO(self.app.get_native_app())

    def get_socket(self) -> SocketIO:
        return self.socket

    def get_test_client(self) -> SocketIOTestClient:
        return self.socket.test_client(self.app.get_native_app())

    def register_handler_cell(self):
        pass

    def emit(
        self,
        event,
        data=None,
        room=None,
        include_self=True,
        namespace=None,
        callback=None) -> None:
        """Emit a custom event to one or more connected clients."""
        return self.socket.emit(
            event, 
            data, 
            room=room,
            include_self=include_self,
            namespace=namespace,
            callback=callback)

    def send(
            self,
            data,
            room=None,
            include_self=True,
            namespace=None,
            callback=None,
            json=False) -> None:
        """Send a message to one or more connected clients."""
        return self.socket.send(
            data,
            room=room,
            include_self=include_self,
            namespace=namespace,
            callback=callback,
            json=json)



import os
import sys
import code
import secrets
from typing import Callable, Type

from flask_cors import CORS
from turbo_flask import Turbo
from flask_session import Session
from flask.testing import FlaskClient
from puft.tools.log import log
from flask import Flask
from flask import cli as flask_cli
from flask.ctx import AppContext, RequestContext
from warepy import get_enum_values
from puft.tools.log import log

from puft.core.sv.sv import Sv
from puft.tools.hints import CLIModeEnumUnion
from puft.core.view.view_cell import ViewCell
from puft.core.cli.cli_run_enum import CLIRunEnum
from .http_method_enum import HTTPMethodEnum
from .turbo_action_enum import TurboActionEnum


class Puft(Sv):
    """Operates over Puft processes.
    
    Should be inherited by project's AppSv.
    """
    def __init__(
            self, 
            config: dict,
            mode_enum: CLIModeEnumUnion,
            host: str,
            port: int,
            ctx_processor_func: Callable | None = None,
            each_request_func: Callable | None = None,
            first_request_func: Callable | None = None) -> None:
        # Templates by default searched within src/app, which allows to
        # integrate them directly to their logical component's folders
        self.DEFAULT_TEMPLATE_PATH = os.path.join(
            config['root_path'], 'src/app')
        self.DEFAULT_STATIC_PATH = os.path.join(
            config['root_path'], 'src/assets')
        self.DEFAULT_INSTANCE_PATH = os.path.join(config['root_path'], 'var')

        # Convert all keys from given config to upper case
        self.config = self._make_upper_keys(config)
        self._assign_defaults_to_config(self.config)
        self._validate_config(self.config)

        super().__init__(self.config)
        self.mode_enum = mode_enum
        self.host = host
        self.port = port
        self.native_app = self._spawn_native_app(self.config)
        self.native_app.config.from_mapping(self.config)
        self._enable_cors(self.config)
        self._enable_testing_config(self.config)
        self._add_random_hex_token_to_config()
        # Initialize turbo.js
        # https://blog.miguelgrinberg.com/post/dynamically-update-your-flask-web-pages-using-turbo-flask
        self.turbo = Turbo(self.native_app)
        # Initialize flask-session after all settings are applied
        self.flask_session = Session(self.native_app)
        self._flush_redis_session_db()

        self._init_app_daemons(
            ctx_processor_func=ctx_processor_func,
            each_request_func=each_request_func,
            first_request_func=first_request_func
        )

    def test_client(self) -> FlaskClient:
        return self.native_app.test_client()

    def _assign_defaults_to_config(self, config: dict) -> None:
        if 'TEMPLATE_PATH' not in config:
            config['TEMPLATE_PATH'] = self.DEFAULT_TEMPLATE_PATH
        if 'STATIC_PATH' not in config:
            config['STATIC_PATH'] = self.DEFAULT_STATIC_PATH
        if 'INSTANCE_PATH' not in config:
            config['INSTANCE_PATH'] = self.DEFAULT_INSTANCE_PATH

    def _validate_config(self, config: dict) -> None:
        """Check if all keys in the config are correct."""
        # TODO
        pass

    def _make_upper_keys(self, mapping: dict) -> dict:
        rmap = {}
        for k, v in mapping.items():
            rmap[k.upper()] = v
        return rmap
    
    def _enable_cors(self, config: dict) -> None:
        # Enable CORS for the app's resources.
        is_cors_enabled = config.get("IS_CORS_ENABLED", None)
        if is_cors_enabled:
            CORS(self.native_app)

    def _flush_redis_session_db(self) -> None:
        # Flush redis session db if mode is not `prod`. 
        if self.mode_enum is not CLIRunEnum.PROD: 
            if self.native_app.config.get("SESSION_TYPE", None):
                if self.native_app.config["SESSION_TYPE"] == "redis":
                    log.info(
                        "Flush session redis Db because"
                        " of non-production run.")
                    session_interface = self.native_app.session_interface
                    redis = session_interface.redis  # type: ignore
                    redis.flushdb()
            else:
                # Apply default null interface, basically do nothing.
                pass
    
    def _add_random_hex_token_to_config(self) -> None:
        # Generate random hex token for App's secret key, if not given in
        # config.
        if self.native_app.config["SECRET_KEY"] is None:
            self.native_app.config["SECRET_KEY"] = secrets.token_hex(16)
            
    def _enable_testing_config(self, config: dict) -> None:
        # Enable testing if appropriate mode has been set. Do not rely on enum
        # if given config explicitly sets TESTING.
        if config.get("TESTING", None) is None:
            if self.mode_enum is CLIRunEnum.TEST:
                self.native_app.config["TESTING"] = True
            else:
                self.native_app.config["TESTING"] = False
        else:
            self.native_app.config["TESTING"] = config["TESTING"]

    def _spawn_native_app(self, config: dict) -> Flask:
        self.instance_path = self.config.get("INSTANCE_PATH")
        self.template_folder = self.config.get("TEMPLATE_PATH") 
        self.static_folder = self.config.get("STATIC_PATH")

        # Set Flask environs according to given mode.
        # Setting exactly through environs instead of config recommended by
        # Flask creators.
        # https://flask.palletsprojects.com/en/2.0.x/config/#:~:text=Using%20the%20environment,a%20previous%20value.
        if self.mode_enum not in [CLIRunEnum.DEV, CLIRunEnum.TEST]:
            os.environ["FLASK_ENV"] = "production"
        else:
            os.environ["FLASK_ENV"] = "development"

        return Flask(
            __name__, 
            instance_path=self.instance_path, 
            template_folder=self.template_folder, 
            static_folder=self.static_folder
        )

    @log.catch
    def get_secret_key(self) -> str:
        """Return secret key defined in App's config."""
        return self.native_app.config["SECRET_KEY"]

    @log.catch
    def get_url(self) -> str:
        """Return app's url.

        Hostname may be e.g. `http://127.0.0.1:5000/`.

        Warning:
            Has url resolving issues described in issue #6.
        """
        return "http://" + self.host + ":" + str(self.port) + "/"

    @log.catch
    def get_native_app(self) -> Flask:
        """Return native app."""
        return self.native_app

    @log.catch
    def get_instance_path(self) -> str:
        """Return app's instance path."""
        return self.native_app.instance_path

    @log.catch
    def run(self) -> None:
        """Run Flask app."""
        self.native_app.run(
            host=self.host, port=self.port
        )

    @log.catch
    def app_context(self) -> AppContext:
        return self.native_app.app_context()

    @log.catch
    def test_request_context(self) -> RequestContext:
        return self.native_app.test_request_context()

    @log.catch
    def register_view(self, view_cell: ViewCell) -> None:
        """Register given view cell for the app."""
        # Check if view has kwargs to avoid sending empty dict.
        # Use cell's name as view's endpoint.
        view = view_cell.view_class.as_view(view_cell.endpoint)
        self.native_app.add_url_rule(
            view_cell.route, view_func=view,
            methods=get_enum_values(HTTPMethodEnum))

    def register_error(
            self,
            error_class: type[Exception],
            handler_function: Callable) -> None:
        self.native_app.register_error_handler(error_class, handler_function)

    @log.catch
    def push_turbo(self, action: TurboActionEnum, target: str, template_path: str, ctx_data: dict = {}) -> None:
        """Push turbo action to target with rendered from path template
        contextualized with given data.
        
        Args:
            action:
                Turbo-Flask action to perform.
            target:
                Id of HTML element to push action to.
            template_path:
                Path to template to render.
            ctx_data (optional):
                Context data to push to rendered template.
                Defaults to empty dict.
        """
        with self.native_app.app_context():
            action = action.value
            target = target
            template_path = template_path
            ctx_data = ctx_data
            exec(
                f"self.turbo.push(self.turbo.{action}"
                "(render_template("
                f"'{template_path}', **{ctx_data}), '{target}'))")

    def postbuild(self) -> None:
        """Abstract method to perform post-injection operation related to app. 
        
        Called by Assembler.

        WARNING:
            This method not working as intended now (runs not actually after
            app starting), so better to not use it temporary.
        
        Raise:
            NotImplementedError:
                If not re-implemented in children.
        """
        raise NotImplementedError(
            "Method `postbuild` hasn't been reimplemented.")

    @log.catch
    def register_cli_cmd(self, *cmds: Callable) -> None:
        """Register cli cmds for the app."""
        # TODO: Implement.
        raise NotImplementedError("CLI cmds support in development.")
        # for cmd in cmds:
        #     self.native_app.cli.add_command(cmd)

    @log.catch
    def register_shell_processor(self, *shell_processors: Callable) -> None:
        """Register shell processors for the app."""
        for processor in shell_processors:
            self.native_app.shell_context_processor(processor)

    @log.catch
    def run_shell(self) -> None:
        """Adapted version of ``Flask.cli.shell_command`` by Pallets.

        Run an interactive Python shell in the context of a given
        Flask application.  The application will populate the default
        namespace of this shell according to its configuration.

        This is useful for executing small snippets of management code
        without having to manually configure the application.
        """
        banner = (
            f"Python {sys.version} on {sys.platform}\n"
            f"App: {self.native_app.import_name} [{self.native_app.env}]\n"
            f"Instance: {self.native_app.instance_path}"
        )
        ctx: dict = {}

        # Support the regular Python interpreter startup script if someone
        # is using it.
        startup = os.environ.get("PYTHONSTARTUP")
        if startup and os.path.isfile(startup):
            with open(startup) as f:
                eval(compile(f.read(), startup, "exec"), ctx)

        ctx.update(self.native_app.make_shell_context())

        # Site, customize, or startup script can set a hook to call when
        # entering interactive mode. The default one sets up readline with
        # tab and history completion.
        interactive_hook = getattr(sys, "__interactivehook__", None)

        if interactive_hook is not None:
            try:
                import readline
                from rlcompleter import Completer
            except ImportError:
                pass
            else:
                # rlcompleter uses __main__.__dict__ by default, which is
                # flask.__main__. Use the shell context instead.
                readline.set_completer(Completer(ctx).complete)

            interactive_hook()

        code.interact(banner=banner, local=ctx)

    def _init_app_daemons(
            self, ctx_processor_func: Callable | None,
            each_request_func: Callable | None,
            first_request_func: Callable | None) -> None:
        """Binds various background processes to the app."""
        flask_app = self.get_native_app()

        if ctx_processor_func:
            @flask_app.context_processor
            def invoke_ctx_processor_func():
                return ctx_processor_func()

        if each_request_func:
            @flask_app.before_request
            def invoke_each_request_func():
                return each_request_func()

        if first_request_func:
            @flask_app.before_first_request
            def invoke_first_request_func():
                return first_request_func()



import json
from abc import abstractmethod
from typing import Any, List, Dict, Literal, Union, Callable, Type

from warepy import Singleton, format_message
from puft.tools.log import log


class Sv(Singleton):
    """Sv superclass inherits Singleton metaclass.

    Layer between Domain Objects (Domains) and UI layer (Controllers and Views).
    Performs various operations e.g. to prepare data (like json formatting) coming to Domains and from them."""
    def __init__(self, config: dict) -> None:
        self.config = config



from dataclasses import dataclass
from typing import Sequence, TypeVar

from puft.core.cell.cell import Cell


# Set TypeVar upper bound to class defined afterwards.
# https://stackoverflow.com/a/67662276
AnyNamedCell = TypeVar("AnyNamedCell", bound="NamedCell")


@dataclass
class NamedCell(Cell):
    name: str

    @staticmethod
    def find_by_name(name: str, cells: Sequence[AnyNamedCell]) -> AnyNamedCell:
        """Traverse through given list of cells and return first one with
        specified name.
        
        Raise:
            ValueError: 
                No cell with given name found.
        """
        for cell in cells:
            if cell.name == name:
                return cell
        raise ValueError(
            "No cell with given name {} found.", name)

    @staticmethod
    def map_to_name(cells: list[AnyNamedCell]) -> dict[str, AnyNamedCell]:
        """Traverse through given cells names and return dict with these cells
        as values and their names as keys."""
        cells_by_name: dict[str, AnyNamedCell] = {}
        for cell in cells:
            cells_by_name[cell.name] = cell
        return cells_by_name


from pytest import fixture
from flask_socketio import SocketIOTestClient
from flask.testing import FlaskClient

from puft.core.app.puft import Puft
from puft.core.db.db import Db
from puft.core.sock.socket import Socket
from puft.tools.get_root_path import get_root_path


class Test:
    @fixture
    def app(self):
        app: Puft = Puft.instance()
        yield app

    @fixture
    def db(self, app: Puft):
        db: Db = Db.instance()

        with app.app_context():
            db.drop_all()
            db.create_all()

        yield db

        with app.app_context():
            db.drop_all()

    @fixture
    def socket(self) -> Socket:
        return Socket.instance()

    @fixture
    def client(self, app: Puft) -> FlaskClient:
        return app.test_client()

    @fixture
    def sock_client(self, app: Puft, socket: Socket) -> SocketIOTestClient:
        # https://github.com/miguelgrinberg/Flask-SocketIO/blob/main/test_socketio.py
        return socket.get_test_client()

    @fixture
    def root_path(self) -> str:
        return get_root_path()



from flask_socketio import SocketIO, Namespace

from .socket import Socket


class Sock(Namespace):
    def __init__(self, namespace=None):
        super().__init__(namespace)
        self.socket: Socket = Socket.instance()

    def on_connect(self):
        self.socket.send('Connected')

    def on_disconnect(self):
        self.socket.send('Disconnected')

    def on_message(self, message):
        self.socket.send(f'Hello {message}!')



from __future__ import annotations
from typing import TYPE_CHECKING
from dataclasses import dataclass

from puft.core.sv.sv_cell import SvCell

if TYPE_CHECKING:
    from puft.core.db.db import Db


@dataclass
class DbSvCell(SvCell):
    """Injection cell with Db itself which can be applied to created
    application.
    """
    sv_class: type[Db]


from puft.core.error.error import Error


def handle_wildcard_error(err: Error):
    return err.expose(), err.status_code


def handle_wildcard_builtin_error(err: Exception):
    return handle_wildcard_error(Error(' ; '.join(err.args), 400))



from puft.core.app.turbo_action_enum import TurboActionEnum
from warepy import Singleton
from puft.tools.log import log

from ..app.puft import Puft


class Emt(Singleton):
    """Emt is a special class performing various tasks related to turbo.js actions, such as updating.
    
    It acts with Views, which call methods from Emts.
    """
    def __init__(self, puft: Puft) -> None:
        self.puft = puft

    def push(
        self, action: TurboActionEnum, target: str, template_path: str, ctx_data: dict = {}  
    ) -> None:
        """Push turbo action to target with rendered from path template contextualized with given data.
        
        Args:
            action: Turbo-Flask action to perform.
            target: Id of HTML element to push action to.
            template_path: Path to template to render.
            ctx_data (optional): Context data to push to rendered template. Defaults to empty dict.
        """
        self.puft.push_turbo(
            action=action,
            target=target,
            template_path=template_path,
            ctx_data=ctx_data
        )



import os
from typing import Callable
from puft.core.sock.default_sock_error_handler import default_sock_error_handler
from puft.core.sock.sock_cell import SockCell

from puft.core.sv.sv_cell import SvCell
from puft.core.view.view_cell import ViewCell
from puft.core.emt.emt_cell import EmtCell
from puft.core.error.error_cell import ErrorCell


class Build:
    """Proxy mapping class with collection of initial project instances to be
    builded by assembler.
    
    Should be inherited by build class in root folder.
    """
    def __init__(
        self,
        version: str = "",
        config_dir: str = "./src/configs",
        sv_cells: list[SvCell] = [],
        view_cells: list[ViewCell] = [],
        emt_cells: list[EmtCell] = [],
        error_cells: list[ErrorCell] = [],
        shell_processors: list[Callable] = [],
        cli_cmds: list[Callable] = [],
        sock_handler_cells: list[SockCell] = [],
        default_sock_error_handler: Callable = default_sock_error_handler,
        ctx_processor_func: Callable | None = None,
        each_request_func: Callable | None = None,
        first_request_func: Callable | None = None,
    ) -> None:
        # Use native root module path.
        # Because this class imported from root directory, root_path below will be assigned from there during class's
        # initialization.
        # NOTE: This path can be overrided by assembler in order of getting arguments on it's own initialization.
        self.root_path = os.getcwd()

        self.version = version
        self.config_dir = config_dir
        self.sv_cells = sv_cells
        self.view_cells = view_cells
        self.emt_cells = emt_cells
        self.error_cells = error_cells
        self.shell_processors = shell_processors
        self.cli_cmds = cli_cmds
        self.ctx_processor_func = ctx_processor_func
        self.each_request_func = each_request_func
        self.first_request_func = first_request_func
        self.sock_handler_cells = sock_handler_cells
        self.default_sock_error_handler = default_sock_error_handler



from typing import Callable
from dataclasses import dataclass

from puft.core.cell.cell import Cell

from .error import Error


@dataclass
class ErrorCell(Cell):
    error_class: type[Error]
    handler_function: Callable


from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable

from puft.core.sv.sv_cell import SvCell

if TYPE_CHECKING:
    from puft.core.app.puft import Puft
    from puft.tools.hints import CLIModeEnumUnion


@dataclass
class PuftSvCell(SvCell):
    """Injection cell with app itself which is required in any build."""
    sv_class: type[Puft]
    mode_enum: CLIModeEnumUnion
    host: str
    port: int
    ctx_processor_func: Callable | None = None
    each_request_func: Callable | None = None
    first_request_func: Callable | None = None


from __future__ import annotations
import re
from functools import wraps
from typing import Callable, Any, TypeVar

from warepy import format_message, snakefy
from puft.core.db.model_not_found_error import ModelNotFoundError
from puft.tools.log import log
from flask import Flask
import flask_migrate
from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy import Model as BaseModel
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declared_attr

from puft.core.sv.sv import Sv
from .db_type_enum import DbTypeEnum


AnyModel = TypeVar('AnyModel', bound='orm.Model')


# TODO: Fix type hinting for decorated functions under this decorator.
def migration_implemented(func: Callable):
    @wraps(func)
    def inner(self_instance, *args, **kwargs):
        if type(self_instance) is not Db:
            raise TypeError(
                "Decorator migration_implemented cannot be"
                f" applied to type {type(self_instance)}")
        elif not self_instance.migration:
            error_message = "Migrate object hasn't been implemented yet"
            raise AttributeError(error_message)
        else:
            return func(self_instance, *args, **kwargs)
    return inner


class Mapper(BaseModel):
    """Base orm model responsible of holding model's data and at least
    it's basic CRUD operations.

    Contains create(), get_first(), get_all() and delete_first() methods as
    Create, Retrieve and Delete representatives.
    Update representatives are defined individually at each subclass
    (e.g. `set_something()`), and by default accessed via basic model
    alteration, e.g. `MyModel.name = 'Another name'`.
    """
    # sqlalchemy used instead of `orm` class to avoid reference errors
    # https://flask-sqlalchemy.palletsprojects.com/en/2.x/customizing/
    id = sa.Column(sa.Integer, primary_key=True)

    @declared_attr
    def __tablename__(cls) -> str:
        cls_name: str = cls.__name__  # type: ignore
        return snakefy(cls_name)

    @declared_attr
    def __mapper_args__(cls) -> dict[str, Any]:
        args: dict[str, Any] = {}
        if hasattr(cls, 'type'):
            # If classes intended to build an inheritance tree, they must
            # include `type` attr
            args.update({
                'polymorphic_on': 'type',
                'polymorphic_identity': cls.__tablename__
            })
        return args

    @classmethod
    @log.catch(exclude=(NotImplementedError))
    def create(cls: AnyModel, **kwargs) -> AnyModel:
        """Create model and return it.
        
        Accepts all given kwargs and thus is recommended to be redefined at
        subclasses.

        Model creation intended to be done only through this method.
        """
        return cls(**kwargs)  # type: ignore

    @classmethod
    @log.catch(exclude=(ModelNotFoundError))
    def get_first(
            cls,
            order_by: object | list[object] | None = None,
            **kwargs) -> orm.Model:
        """Filter first ORM mapped model by given kwargs and return it.
        
        Raise:
            ValueError:
                No such ORM model in db matched given kwargs
        """
        query: Any = cls.query.filter_by(**kwargs)  # type: ignore

        if order_by is not None:
            query = cls._order_query(query, order_by)

        model: orm.Model = query.first()

        if not model:
            raise ModelNotFoundError(model_name=cls.__name__, **kwargs)
        else:
            return model

    @classmethod
    @log.catch(exclude=(ModelNotFoundError))
    def get_all(
            cls,
            order_by: object | list[object] | None = None,
            limit: int | None = None,
            **kwargs) -> list[orm.Model]:
        """Filter all ORM mapped models by given kwargs and return them.

        Raise:
            ValueError:
                No such ORM model in db matched given kwargs
        """
        query: Any = cls.query.filter_by(**kwargs)  # type: ignore

        if order_by is not None:
            query = cls._order_query(query, order_by)

            if limit:
                query = query.limit(limit)
        elif limit:
            query = query.limit(limit)

        models: list[orm.Model] = query.all()

        if not models:
            raise ModelNotFoundError(model_name=cls.__name__, **kwargs)
        else:
            return models

    @classmethod
    def delete_first(
            cls,
            order_by: object | list[object] | None = None,
            **kwargs) -> None:
        """Delete first accessed by `get_first()` method model."""
        db: Db = Db.instance()
        model: orm.Model = cls.get_first(order_by=order_by, **kwargs)

        db.native_db.session.delete(model)
        db.commit()

    @staticmethod
    def _order_query(query: Any, order_by: object | list[object]) -> object:
        if type(order_by) is list:
            return query.order_by(*order_by)
        else:
            return query.order_by(order_by)


class orm:
    # Helper references for shorter writing at ORMs.
    # Ignore lines added for a workaround to fix issue:
    # https://github.com/microsoft/pylance-release/issues/187
    native_db = SQLAlchemy(model_class=Mapper)
    Model: Any = native_db.Model 
    column = native_db.Column  # type: ignore 
    integer = native_db.Integer  # type: ignore
    string = native_db.String  # type: ignore
    text = native_db.Text  # type: ignore
    float = native_db.Float  # type: ignore
    boolean = native_db.Boolean  # type: ignore
    foreign_key = native_db.ForeignKey  # type: ignore
    table = native_db.Table  # type: ignore
    check_constraint = native_db.CheckConstraint  # type: ignore
    relationship = native_db.relationship  # type: ignore
    backref = native_db.backref  # type: ignore
    pickle = native_db.PickleType  # type: ignore
    binary = native_db.LargeBinary  # type: ignore
    datetime = native_db.DateTime  # type: ignore


class Db(Sv):
    """Operates over database processes."""
    def __init__(self, config: dict) -> None:
        super().__init__(config)
        self.DEFAULT_URI = f"sqlite:///{self.config['root_path']}/sqlite3.db"

        self.native_db = orm.native_db
        # For now sv config propagated to Db domain.
        self._assign_uri_from_config(config)

    def _assign_uri_from_config(self, config: dict) -> None:
        raw_uri = config.get("uri", None)  # type: str

        if not raw_uri:
            log.info(f"URI for database is not specified, using default")
            raw_uri = self.DEFAULT_URI
        else:
            # Case 1: SQLite Db.
            # Developer can give relative path to the Db (it will be absolutized at ConfigCell.parse()),
            # by setting sqlite Db extension to `.db`, e.g. `./instance/sqlite3.db`,
            # or by setting full absolute path with protocol, e.g. `sqlite:////home/user/project/instance/sqlite3.db`.
            if raw_uri.rfind(".db") != -1 or "sqlite:///" in raw_uri:
                if "sqlite:///" not in raw_uri: 
                    # Set absolute path to db.
                    # Ref: https://stackoverflow.com/a/44687471/14748231
                    self.uri = "sqlite:///" + raw_uri
                else:
                    self.uri = raw_uri
                self.type_enum = DbTypeEnum.SQLITE
            # Case 2: PostgreSQL Db.
            elif re.match(r"postgresql(\+\w+)?://", raw_uri):
                # No need to calculate path since psql uri should be given in full form.
                self.uri = raw_uri
                self.type_enum = DbTypeEnum.PSQL
            else:
                raise ValueError(format_message("Unrecognized or yet unsupported type of Db uri: {}", raw_uri))
            
            # WARNING: Never print full Db uri to config, since it may contain user's password (as in case of
            # psql)
            log.info(f"Set db type: {self.type_enum.value}")

    @migration_implemented
    def init_migration(self, directory: str = "migrations", multidb: bool = False) -> None:
        """Initializes migration support for the application."""
        flask_migrate.init(directory=directory, multidb=multidb)

    @migration_implemented
    def migrate_migration(
        self,
        directory: str = "migrations", 
        message = None, 
        sql = False, 
        head: str = "head", 
        splice = False, 
        branch_label = None, 
        version_path = None, 
        rev_id = None
    ) -> None:
        flask_migrate.migrate(
            directory=directory,
            message=message,
            sql=sql,
            head=head,
            splice=splice,
            branch_label=branch_label,
            version_path=version_path,
            rev_id=rev_id
        )

    @migration_implemented
    def upgrade_migration(
        self,
        directory: str = "migrations",
        revision: str = "head",
        sql = False,
        tag = None
    ) -> None:
        flask_migrate.upgrade(
            directory=directory,
            revision=revision,
            sql=sql,
            tag=tag
        )

    def setup(self, flask_app: Flask) -> None:
        """Setup Db and migration object with given Flask app."""
        flask_app.config["SQLALCHEMY_DATABASE_URI"] = self.uri
        flask_app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
        self.native_db.init_app(flask_app)

        # render_as_batch kwarg required only for sqlite3 databases to avoid
        # ALTER TABLE issue on migrations
        # https://blog.miguelgrinberg.com/post/fixing-alter-table-errors-with-flask-migrate-and-sqlite
        if self.type_enum is DbTypeEnum.SQLITE:
            is_sqlite_db = True
        else:
            is_sqlite_db = False
        self.migration = flask_migrate.Migrate(flask_app, self.native_db, render_as_batch=is_sqlite_db)

    def get_native_db(self) -> SQLAlchemy:
        return self.native_db

    @migration_implemented
    def get_migration(self) -> flask_migrate.Migrate:
        """Return migration object.
        
        Raise:
            AttributeError: If Migrate object hasn't implemented yet.
        """
        return self.migration

    @migration_implemented
    def create_all(self) -> None:
        self.native_db.create_all()

    @migration_implemented
    def drop_all(self):
        "Drop all tables."
        self.native_db.drop_all()

    @migration_implemented
    def _add(self, entity):
        """Place an object in the session."""
        # TODO: Add functionality to accept multiple entities as *args.
        self.native_db.session.add(entity)

    def push(self, entity):
        """Add entity to session and immediately commit the session."""
        self._add(entity)
        self.commit()

    @migration_implemented
    def commit(self):
        """Commit current transaction."""
        self.native_db.session.commit()

    @migration_implemented
    def rollback(self):
        self.native_db.session.rollback()

    @migration_implemented
    def remove(self):
        self.native_db.session.remove()



from __future__ import annotations
from typing import TYPE_CHECKING
from dataclasses import dataclass

from puft.core.sv.sv_cell import SvCell

if TYPE_CHECKING:
    from puft.core.sock.socket import Socket


@dataclass
class SocketSvCell(SvCell):
    sv_class: type[Sock]
